from docker_replay.models import DockerArg

class ArgParser(object):
    def __init__(self, o_name, o_key):
        self.name = o_name
        self.key = o_key

    def build(self, val):
        yield DockerArg(self.name, val)

class CmdParser(ArgParser):
    def build(self, val):
        if val:
            val = ' '.join(val)
        yield DockerArg(self.name, val)

config_args = [
  ArgParser('image', 'Config.Image'),
  CmdParser('cmd', 'Config.Cmd'),
]



#TODO: ulimits

from docker_replay.models import *

class OptParser(object):
    def __init__(self, o_name, o_key, o_type):
        self.name = o_name
        self.key = o_key
        self.otype = o_type

    # parse given value and yield zero or more DockerOpts
    def build(self, val):
        # yield a new opt for multi-value options(--volume, --env, etc.)
        if self.otype == ValueOpt and isinstance(val, list):
            for v in val:
                yield self.build_one(v)
        elif self.otype == MapOpt and isinstance(val, dict):
            for k,v in val.items():
                val = '%s=%s' % (k,v)
                yield self.build_one(val)
        else:
            yield self.build_one(val)

    def build_one(self, val):
        return self.otype(self.name, val)

class NameParser(OptParser):
    def build(self, val):
        yield self.build_one(val.strip('/'))

class PublishedParser(OptParser):
    def build(self, val):
        if not val:
            return

        def read_hostports(hplist):
            for hp in hplist:
                if hp['HostIp']:
                    yield '%s:%s' % (hp['HostIp'], hp['HostPort'])
                else:
                    yield hp['HostPort']

        for cport, binds in val.items():
            cport, proto = cport.split('/')
            for hostport in read_hostports(binds):
                v = '%s:%s/%s' % (hostport,cport,proto)
                yield self.build_one(v)

class ExposedParser(OptParser):
    def build(self, val):
        if not val:
            return
        for port in val.keys():
            yield self.build_one(port)

class LinkParser(OptParser):
    def build(self, val):
        if not val:
            return
        for link in val:
            src, linkname = link.split(':')
            v = '%s:%s' % (src.strip('/'), linkname.split('/')[-1])
            yield self.build_one(v)

class NetModeParser(OptParser):
    def build(self, val):
        if val == "default":
            return
        yield self.build_one(val)

class RestartParser(OptParser):
    def build(self, val):
        if val['Name'] == 'no':
            return
        v = val['Name']
        if val['MaximumRetryCount'] != 0:
            v += ':%s' % val['MaximumRetryCount']
        yield self.build_one(v)

class EntrypointParser(OptParser):
    def build(self, val):
        if val:
            v = '"%s"' % ' '.join(val)
            yield self.build_one(v)

config_opts = [
    OptParser('--env', 'Config.Env', ValueOpt),
    OptParser('--hostname', 'Config.Hostname', ValueOpt),
    OptParser('--interactive', 'Config.OpenStdin', BoolOpt),
    OptParser('--label', 'Config.Labels', MapOpt),
    OptParser('--tty', 'Config.Tty', BoolOpt),
    OptParser('--user', 'Config.User', ValueOpt),
    OptParser('--workdir', 'Config.WorkingDir', ValueOpt),
    OptParser('--add-host', 'HostConfig.ExtraHosts', ValueOpt),
    OptParser('--blkio-weight', 'HostConfig.BlkioWeight', ValueOpt),
    OptParser('--blkio-weight-device', 'HostConfig.BlkioWeightDevice', ValueOpt),
    OptParser('--cap-add', 'HostConfig.CapAdd', ValueOpt),
    OptParser('--cap-drop', 'HostConfig.CapDrop', ValueOpt),
    OptParser('--cgroup-parent', 'HostConfig.CgroupParent', ValueOpt),
    OptParser('--cidfile', 'HostConfig.ContainerIDFile', ValueOpt),
    OptParser('--cpu-period', 'HostConfig.CpuPeriod', ValueOpt),
    OptParser('--cpu-shares', 'HostConfig.CpuShares', ValueOpt),
    OptParser('--cpu-quota', 'HostConfig.CpuQuota', ValueOpt),
    OptParser('--cpuset-cpus', 'HostConfig.CpusetCpus', ValueOpt),
    OptParser('--cpuset-mems', 'HostConfig.CpusetMems', ValueOpt),
    OptParser('--device', 'HostConfig.Devices', ValueOpt),
    OptParser('--device-read-bps', 'HostConfig.BlkioDeviceReadBps', ValueOpt),
    OptParser('--device-read-iops', 'HostConfig.BlkioDeviceReadIOps', ValueOpt),
    OptParser('--device-write-bps', 'HostConfig.BlkioDeviceWriteBps', ValueOpt),
    OptParser('--device-write-iops', 'HostConfig.BlkioDeviceWriteIOps', ValueOpt),
    OptParser('--dns', 'HostConfig.Dns', ValueOpt),
    OptParser('--dns-opt', 'HostConfig.DnsOptions', ValueOpt),
    OptParser('--dns-search', 'HostConfig.DnsSearch', ValueOpt),
    OptParser('--group-add', 'HostConfig.GroupAdd', ValueOpt),
    OptParser('--ipc', 'HostConfig.IpcMode', ValueOpt),
    OptParser('--isolation', 'HostConfig.Isolation', ValueOpt),
    OptParser('--kernel-memory', 'HostConfig.KernelMemory', ValueOpt),
    OptParser('--log-driver', 'HostConfig.LogConfig.Type', ValueOpt),
    OptParser('--log-opt', 'HostConfig.LogConfig.Config', MapOpt),
    OptParser('--memory', 'HostConfig.Memory', ByteValueOpt),
    OptParser('--memory-reservation', 'HostConfig.MemoryReservation', ByteValueOpt),
    OptParser('--memory-swap', 'HostConfig.MemorySwap', ByteValueOpt),
    OptParser('--memory-swappiness', 'HostConfig.MemorySwappiness', ValueOpt),
    OptParser('--oom-kill-disable', 'HostConfig.OomKillDisable', BoolOpt),
    OptParser('--oom-score-adj', 'HostConfig.OomScoreAdj', ValueOpt),
    OptParser('--publish-all', 'HostConfig.PublishAllPorts', BoolOpt),
    OptParser('--pid', 'HostConfig.PidMode', ValueOpt),
    OptParser('--pids-limit', 'HostConfig.PidsLimit', ValueOpt),
    OptParser('--privileged', 'HostConfig.Privileged', BoolOpt),
    OptParser('--read-only', 'HostConfig.ReadonlyRootfs', BoolOpt),
    OptParser('--rm', 'HostConfig.AutoRemove', BoolOpt),
    OptParser('--volume', 'HostConfig.Binds', ValueOpt),
    OptParser('--security-opt', 'HostConfig.SecurityOpt', ValueOpt),
    OptParser('--shm-size', 'HostConfig.ShmSize', ByteValueOpt),
    OptParser('--userns', 'HostConfig.UsernsMode', ValueOpt),
    OptParser('--uts', 'HostConfig.UTSMode', ValueOpt),
    OptParser('--volume-driver', 'HostConfig.VolumeDriver', ValueOpt),
    OptParser('--volumes-from', 'HostConfig.VolumesFrom', ValueOpt),
    OptParser('--ip', 'NetworkSettings.IPAddress', ValueOpt),
#    OptParser('--ip6', 'LinkLocalIPv6Address', ValueOpt),
#    OptParser('--netdefault', '???', ValueOpt),
#    OptParser('--net-alias', '???', ValueOpt),
    OptParser('--mac-address', 'NetworkSettings.MacAddress', ValueOpt),

    # non-generic opt parsers
    EntrypointParser('--entrypoint', 'Config.Entrypoint', DockerOpt),
    ExposedParser('--expose', 'Config.ExposedPorts', DockerOpt),
    LinkParser('--link', 'HostConfig.Links', DockerOpt),
    NameParser('--name', 'Name', DockerOpt),
    NetModeParser('--net', 'HostConfig.NetworkMode', DockerOpt),
    PublishedParser('--publish', 'HostConfig.PortBindings', DockerOpt),
    RestartParser('--restart', 'HostConfig.RestartPolicy', DockerOpt),
  ]




class DockerParam(object):
    """ Base class for options or arguments """

    default = None

    def __init__(self, name, val):
        self.name = name
        self.value = val

    def is_null(self):
        if self.value:
            return False
        return True

class DockerArg(DockerParam):
    """ Represents a positional argument to `docker run` """

    def __str__(self):
        return self.value

class DockerOpt(DockerParam):
    """ Represents an option to `docker run` """

    def __str__(self):
        return '%s %s' % (self.name, self.value)

"""
Generic option types
"""

class BoolOpt(DockerOpt):
    def __init__(self, *args):
        super(DockerOpt, self).__init__(*args)
        # DockerOpt.__init__(self, *args)

    def __str__(self):
        return self.name

class ByteValueOpt(DockerOpt):
    """ Option with one or more user-defined values """

    @staticmethod
    def format_bytes(x):
        KB = 1024
        MB = KB*1024
        GB = MB*1024

        def _round(x):
            return int(round(x))

        x = float(x)
        if x < 1024:
            return '%sb' % _round(x)
        elif 1024 <= x < MB:
            return '%sk' % _round(x/KB)
        elif KB <= x < GB:
            return '%sm' % _round(x/MB)
        elif GB <= x:
            return '%sg' % _round(x/GB)

    def __str__(self):
        try:
            return '%s %s' % (self.name, self.format_bytes(self.value))
        except ValueError:
            raise TypeError('unsupported value type for option "%s": %s' % \
                    (self.name, self.value))

class ValueOpt(DockerOpt):
    """ Option with one or more user-defined values """

    def __str__(self):
        try:
            return '%s %s' % (self.name, self.value)
        except ValueError:
            raise TypeError('unsupported value type for option "%s": %s' % \
                    (self.name, self.value))

class MapOpt(ValueOpt):
    """ Option with one or more user-defined mappings """

    default = {}



import os
import sys
import logging
from argparse import ArgumentParser

from docker_replay.version import version

log = logging.getLogger('docker-replay')

class DockerReplay(object):
    def __init__(self, container_id, pretty_print=True):
        from docker import client, errors
        from docker_replay.parser import ConfigParser

        self.pretty_print = pretty_print

        c = client.from_env(version='auto')

        try:
            inspect = c.api.inspect_container(container_id)
            self.parser = ConfigParser(inspect)
        except errors.NotFound:
            print('no such container: %s' % container_id)
            sys.exit(1)

    def __str__(self):
        opts = sorted([ str(o) for o in self.parser.opts if not o.is_null() ])
        opts += [ str(a) for a in self.parser.args if not a.is_null() ]

        if self.pretty_print:
            return 'docker run %s' % ' \\\n           '.join(opts)
        return 'docker run %s' % ' '.join(opts)

def main():
    argparser = ArgumentParser(description='docker-replay v%s' % version)
    argparser.add_argument('-d', '--debug', action='store_true',
                            help='enable debug output')
    argparser.add_argument('-p', '--pretty-print', action='store_true',
                            help='pretty-print output')
    argparser.add_argument('container',
                            help='container to generate command from')
    args = argparser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.WARN)

    print(DockerReplay(args.container, args.pretty_print))



from setuptools import setup

exec(open('docker_replay/version.py').read())

setup(name='docker-replay',
      version=version,
      packages=['docker_replay'],
      description='Generate docker run commands from running containers',
      author='Bradley Cicenas',
      author_email='bradley.cicenas@gmail.com',
      url='https://github.com/bcicen/docker-replay',
      install_requires=['docker>=2.4.2'],
      license='http://opensource.org/licenses/MIT',
      classifiers=(
          'Intended Audience :: Developers',
          'License :: OSI Approved :: MIT License ',
          'Natural Language :: English',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3.4',
      ),
      keywords='docker docker-py devops',
      entry_points = {
        'console_scripts' : ['docker-replay = docker_replay:main']
      }
)



import logging
from docker_replay.args import config_args
from docker_replay.opts import config_opts
from docker_replay.models import DockerOpt, DockerArg

log = logging.getLogger('docker-replay')

class ConfigParser(object):

    args = []
    opts = []

    def __init__(self, config):
        self.config = config

        # build all options
        for op in config_opts:
            o_val = self.get(op.key, op.otype.default)
            self.opts += list(op.build(o_val))

        for ap in config_args:
            o_val = self.get(ap.key)
            self.args += list(ap.build(o_val))

        log.info('parsed %d options' % len(self.opts))
        log.info('parsed %d args' % len(self.args))

    def get(self, key, default=None):
        """
        Retrieve a top-level or nested key, e.g:
        >>> get('Id')
        >>> get('HostConfig.Binds')
        """
        key_parts = key.split('.')
        config = self.config
        while key_parts:
            try:
                config = config[key_parts.pop(0)]
            except KeyError:
                log.warn('returning default for missing key: %s' % key)
                return default
        log.debug('get key: %s (%s)' % (key, config))
        return config
